from rc import run, running, run_stream, handle_stream, STDERR, STDOUT, EXIT, print_stream, save_stream_to_file, gcloud
import contextlib
from io import StringIO
import time


def test_run_sh_syntax():
    assert run(['ls', '~']).returncode == 0
    assert run('ls ~').returncode == 0
    assert run(['cat', '"~"']) == run('cat "~"')


def test_run_multiple_line():
    assert run('cat', input='''hello world
aaa''').stdout == 'hello world\naaa'

    assert run('bash', input='''date
ls /
echo hello world
''').stdout.strip().split('\n')[-1] == 'hello world'

    assert run('python', input='''
print('hello world')
print('aaa')
''').stdout.strip() == 'hello world\naaa'


def test_run_stream():
    q, _ = run_stream('python', input='''
import sys
import time
sys.stdout.write('to stdout\\n')
sys.stdout.flush()
time.sleep(1) # sleep to make stdout happens earlier than stderr for sure
sys.stderr.write("to stderr\\n")
sys.stderr.flush()
time.sleep(1)
sys.stdout.write('hello world\\n')
sys.stdout.flush()
exit(1)
    ''')

    t = q.get()
    assert t == (STDOUT, 'to stdout\n')
    q.task_done()

    t = q.get()
    assert t == (STDERR, 'to stderr\n')
    q.task_done()

    t = q.get()
    assert t == (STDOUT, 'hello world\n')
    q.task_done()

    t = q.get()
    assert t == (EXIT, 1)
    q.task_done()

    assert q.empty()


def test_handle_stream():
    q, _ = run_stream('sh', input='''
    echo to stdout
    echo 1>&2 to stderr
    echo hello world
    exit 1
    ''')
    stdout = []
    stderr = []
    exitcode = []
    handle_stream(q, stdout_handler=lambda line: stdout.append(
        line), stderr_handler=lambda line: stderr.append(line), exit_handler=lambda x: exitcode.append(x))
    assert stdout == ['to stdout\n', 'hello world\n']
    assert stderr == ['to stderr\n']
    assert exitcode == [1]


def test_print_stream():
    q, _ = run_stream('sh', input='''
    echo to stdout
    echo 1>&2 to stderr
    echo hello world
    exit 1
    ''')
    temp_stdout = StringIO()
    with contextlib.redirect_stdout(temp_stdout):
        print_stream(q, prefix='node1')

    output = temp_stdout.getvalue().split('\n')
    assert output[-1] == 'node1 EXIT CODE | 1'
    assert output.sort() == """node1 STDERR | to stderr
node1 STDOUT | to stdout
node1 STDOUT | hello world
node1 EXIT CODE | 1""".split('\n').sort()


def save_stream_to_file():
    q, _ = run_stream('sh', input='''
    echo to stdout
    echo 1>&2 to stderr
    echo hello world
    exit 1
    ''')
    save_stream_to_file(path='/tmp', name='node1')
    assert open('/tmp/node1.stdout').read() == '''node1 STDOUT | to stdout
node1 STDOUT | hello world'''
    assert open('/tmp/node1.stderr').read() == 'node1 STDERR | to stderr\n'
    assert open('/tmp/node1.exitcode').read() == 'node1 EXIT CODE | 1'


def test_tmux():
    machine = gcloud.create(name="test-rc-node", machine_type="n1-standard-1", disk_size="20G", image_project='ubuntu-os-cloud', image_family='ubuntu-1804-lts',
                            zone='us-west2-a', preemptible=False, firewall_allows=['tcp:8080'])
    machine = gcloud.get('test-rc-node')
    machine.run('sudo apt install tmux')
    machine.run_detach_tmux('while true; do echo aaaaaa; sleep 10; done')
    machine.kill_detach_tmux()
    assert machine.run('cat /tmp/python-rc.log').stdout == 'aaaaaa\n'
    machine.delete()



from rc.provider import gcloud, azure, digitalocean
from rc.machine import Machine
from rc.util import run, RunException, RunResult, running, run_stream, handle_stream, \
    STDERR, STDOUT, EXIT, go, pmap, as_completed, print_stream, save_stream_to_file, \
    p, ep, bash, python, python2, python3, sudo, kill, ok



from rc.util import run
import json
from functools import lru_cache


@lru_cache(maxsize=1)
def _get_regions():
    regions = json.loads(run(
        ['aws', 'ec2', 'describe-regions', '--region', 'us-west-1']).stdout)['Regions']
    return list(map(lambda region: region['RegionName'], regions))


def list():
    regions = _get_regions()
    instances = []
    for r in regions:
        region_instances = json.loads(run(
            ['aws', 'ec2', 'describe-instances', '--region', r]).stdout)['Reservations'][0]



from rc.util import run
from rc.exception import MachineCreationException, MachineDeletionException, \
    MachineShutdownException, MachineBootupException, SaveImageException, MachineChangeTypeException, \
    DeleteImageException, FirewallRuleCreationException
from rc.machine import Machine
from rc.firewall import Firewall
import sys
import re
import os
from functools import lru_cache
import json
import time

digitalocean_provider = sys.modules[__name__]

SSH_KEY_PATH = os.path.expanduser('~/.ssh/id_rsa')


@lru_cache(maxsize=1)
def _digitalocean_ssh_key_fingerprint():
    p = run('ssh-keygen -E md5 -lf ~/.ssh/id_rsa.pub')
    fingerprint = p.stdout.split(' ')[1][4:]
    p = run(f'doctl compute ssh-key get {fingerprint}')
    if p.returncode != 0:
        p = run(f'doctl compute ssh-key import python-rc ~/.ssh/id_rsa.pub')
    return fingerprint


def list():
    p = run(['doctl', 'compute', 'droplet', 'list', '--no-header',
             '--format', 'Region,Name,PublicIPv4,ID'])
    result = []
    lines = p.stdout.strip('\n').split('\n')
    for line in lines:
        zone, name, ip, id_ = re.split(r'\s+', line)
        m = Machine(provider=digitalocean_provider, name=name,
                    zone=zone, ip=ip, username='root', ssh_key_path=SSH_KEY_PATH)
        m.id = id_
        result.append(m)
    return result


def _exist(id_):
    p = run(f'doctl compute droplet get {id_}')
    return p.returncode == 0


def get(name):
    p = run(
        f'doctl compute droplet list --no-header --format Region,Name,PublicIPv4,ID')
    lines = p.stdout.strip('\n').split('\n')
    for line in lines:
        zone, name_, ip, id_ = re.split(r'\s+', line)
        if name_ == name:
            m = Machine(provider=digitalocean_provider, name=name,
                        zone=zone, ip=ip, username='root', ssh_key_path=SSH_KEY_PATH)
            m.id = id_
            return m
    return None


def status(machine):
    p = run(
        f'doctl compute droplet get {machine.id} --no-header --format Status')
    return p.stdout.strip()


def bootup(machine):
    p = run(f'doctl compute droplet-action power-on {machine.id} --wait')
    if p.returncode != 0:
        return MachineBootupException(p.stderr)
    machine.wait_ssh()


def shutdown(machine):
    p = run(f'doctl compute droplet-action shutdown {machine.id} --wait')
    if p.returncode != 0:
        raise MachineShutdownException(p.stderr)


def create(name, *, image, region, size, firewall_names=None):
    # Available images:
    # user images: doctl compute snapshot list
    # digitalocean linux distro images: doctl compute image list-distribution
    # digitalocean application images: doctl compute image list-application
    # Use the slug for digitalocean images. Use name for user images

    # Available regions:
    # doctl compute region list
    # Use slug to refer a region

    # Available machine sizes:
    # doctl compute size list
    # Use slug to refer a machine size
    machine = get(name)
    if machine:
        raise MachineCreationException(f'Machine {name} is already exist')
    cmd = f'doctl compute droplet create {name} --region {region} --size {size} --image {image} --ssh-keys {_digitalocean_ssh_key_fingerprint()}'
    if firewall_names:
        cmd += ' --tag-names ' + ','.join(firewall_names)
    cmd += ' --wait'
    p = run(cmd)
    if p.returncode != 0:
        raise MachineCreationException(p.stderr)
    machine = get(name)
    machine.wait_ssh()
    return machine


def change_type(machine, new_type):
    # new_type: a digitalocean machine size
    # doctl compute size list
    # Use slug to refer a machine size
    p = run(
        f'doctl compute droplet-action resize {machine.id} --size {new_type} --wait')
    if p.returncode != 0:
        raise MachineChangeTypeException(p.stderr)


def save_image(machine, image):
    p = run(
        f'doctl compute droplet-action snapshot {machine.id} --snapshot-name {image} --wait')
    if p.returncode != 0:
        raise SaveImageException(p.stderr)


def delete_image(image):
    p = run(
        f'doctl compute snapshot list --output json')
    snapshots = json.load(p.stdout)
    for s in snapshots:
        if s["name"] == image:
            p = run(f'doctl compute snapshot delete {s["id"]}')
            if p.returncode != 0:
                raise DeleteImageException(p.stderr)
            return


def delete(machine):
    p = run(f'doctl compute droplet delete {machine.id} --force')
    if p.returncode != 0:
        raise MachineDeletionException(p.stderr)
    while _exist(machine):
        time.sleep(1)


def create_firewall(name, *, direction='in', ports, ips=['0.0.0.0/0']):
    cmd = f'doctl compute firewall create {name} --tag-names {name}'
    rules = []
    for port in ports:
        if port == 'icmp':
            rule = 'protocol:icmp,address:'
            rule += ',address:'.join(ips)
        else:
            protocol, port = port.split(':')
            rule = f'protocol:{prototocol},port:{port},address:'
            rule += ',address:'.join(ips)
        rules.append(rule)
    rules = ' '.join(rules)
    if direction == 'in':
        cmd += f" --inbound-rules '{rules}'"
    elif direction == 'out':
        cmd += f" --outbound-rules '{rules}'"
    else:
        raise FirewallRuleCreationException(
            'direction must be either "in" or "out"')
    p = run(cmd)
    if p.returncode != 0:
        raise FirewallRuleCreationException(p.stderr)
    return Firewall(name, provider=digitalocean_provider, direction=direction, action=action, ports=ports, ips=ips)



from rc.test.util import Timer
from rc import gcloud
from rc.util import go, pmap, as_completed


def create_or_get(*, name, **kwargs):
    return gcloud.get(name) or gcloud.create(name=name, **kwargs)


def test_create_delete_5_instance():
    futures = []
    machines = []
    with Timer('create 5 instances'):
        for i in range(5):
            futures.append(go(create_or_get, name="test-rc-node-"+str(i), machine_type="n1-standard-1", disk_size="20G", image_project='ubuntu-os-cloud', image_family='ubuntu-1804-lts',
                              zone='us-west2-a', preemptible=False, firewall_allows=['tcp:8080']))
        for f in as_completed(futures):
            m = f.result()
            print("Created machine:", m.name)
            machines.append(m)

    futures = {}

    def delete_print(m):
        m.delete()
        print("Deleted machine:", m.name)
    with Timer('delete 5 instances'):
        pmap(delete_print, machines)

    for m in machines:
        assert gcloud.get(m.name) is None



import os
from rc import run, ok, pmap, p
import yaml
import rc

config_dir = os.path.expanduser('~/.python-rc/')
groups_dir = os.path.join(config_dir, 'groups')
logs_dir = os.path.join(config_dir, 'logs')


def create_config_dirs():
    ok(run(f'mkdir -p {groups_dir}'))
    ok(run(f'mkdir -p {logs_dir}'))
    with open(os.path.join(config_dir, '.tmux.conf'), 'w') as f:
        f.write('''set -g mouse on
bind-key a set-window-option synchronize-panes
''')


def get_spec(spec, default, item):
    if type(spec) is str:
        return default.get(item)
    return spec.get(item, default.get(item))


def machine_from_spec(spec, default=None):
    if default is None:
        default = {}
    provider = get_spec(spec, default, 'provider')
    assert provider
    provider = getattr(rc.provider, provider)
    assert provider
    if type(spec) is str:
        machine = provider.get(spec, **default)
        assert machine
        return machine
    else:
        name = spec.get('name')
        spec = {**default, **spec}
        machine = provider.get(name, **spec)
        assert machine
        return machine


def get_targets(arg):
    group_ = arg.split('/')
    group_config_file = get(group_[0])
    if group_config_file:
        try:
            if len(group_) > 1:
                sub_machines = group_[1].split(',')
            else:
                sub_machines = None
            return parse_config(group_config_file, sub_machines)
        except Exception as e:
            print(f'targets: {arg} invalid')
            print(e)
            exit(2)
    else:
        return None


def parse_config(config_file, sub_machines=None):
    config = yaml.safe_load(open(config_file))
    default = config.get('default', {})
    machine_spec = config.get('machines', None)
    if machine_spec is None:
        raise "Machines cannot be empty"
    machines = pmap(lambda ms: machine_from_spec(ms, default), machine_spec)
    group_machines = {}
    for m in machines:
        group_machines[str(m)] = m
    assert len(group_machines) == len(machines)
    if sub_machines:
        targets = []
        for n in sub_machines:
            matches = list(
                filter(lambda name: n in name, group_machines.keys()))
            assert len(matches) == 1
            targets += matches
        assert len(targets) > 0
    else:
        targets = group_machines.keys()
    targets = sorted(targets)
    p('machine targets: ', ', '.join(targets))
    return list(map(lambda n: group_machines[n], targets))


def get(group_name):
    if os.path.exists(os.path.join(groups_dir, group_name)):
        return os.path.join(groups_dir, group_name)



from rc import gcloud
from rc.util import run
from rc.test.util import Timer


def create_test_machine():
    return gcloud.create(name="test-rc-node", machine_type="n1-standard-1", disk_size="20G", image_project='ubuntu-os-cloud', image_family='ubuntu-1804-lts',
                         zone='us-west2-a', preemptible=False, firewall_allows=['tcp:8080'])


def test_gcloud():
    # Cleanup from last fail
    a = gcloud.get('test-rc-node')
    if a is not None:
        a.delete()

    # Test list
    with Timer('list machine'):
        old_machines = gcloud.list()
    assert len(list(filter(lambda machine: machine.name ==
                           'test-rc-node', old_machines))) == 0

    # Test get None
    assert gcloud.get('test-rc-node') is None

    # Test create machine
    with Timer('create machine'):
        machine1 = create_test_machine()
    assert machine1.name == "test-rc-node"
    assert machine1.ip != ''
    assert machine1.ssh_key_path != ''
    assert machine1.username != ''
    assert machine1.status() == "RUNNING"

    # Test get machine
    with Timer('get machine'):
        machine1_get = gcloud.get('test-rc-node')
    assert machine1 == machine1_get

    # Test list after create
    new_machines = gcloud.list()
    assert len(list(filter(lambda machine: machine.name ==
                           'test-rc-node', new_machines))) == 1
    assert len(new_machines) == len(old_machines) + 1

    # Test shutdown
    with Timer('shutdown'):
        machine1.shutdown()
    assert machine1.status() == "TERMINATED"

    # Test bootup
    with Timer('bootup'):
        machine1.bootup()
    assert machine1.status() == "RUNNING"

    # Test upload download
    machine1.upload(local_path='rc/test/test_gcloud.py', machine_path="/tmp/")
    assert machine1.run("ls /tmp/test_gcloud.py")
    machine1.download(machine_path="/tmp/test_gcloud.py",
                      local_path="/tmp/test_gcloud.py")
    assert open(
        "/tmp/test_gcloud.py").read() == open("/tmp/test_gcloud.py").read()
    run(['rm', '/tmp/test_gcloud.py'])

    # Test run command on machine
    machine1.run(['echo', 'aaa', '>', '/tmp/aaaa.txt'])
    assert run('ls /tmp/aaaa.txt').returncode != 0
    machine1.run(['grep', '', '/tmp/aaaa.txt'])
    p = machine1.run('cat /tmp/aaaa.txt')
    assert p.stdout == 'aaa\n'

    # Test delete
    with Timer('delete machine'):
        machine1.delete()
    assert gcloud.get('test-rc-node') is None



class RcException(Exception):
    pass


class MachineCreationException(RcException):
    pass


class MachineDeletionException(RcException):
    pass


class MachineBootupException(RcException):
    pass


class MachineShutdownException(RcException):
    pass


class UploadException(RcException):
    pass


class DownloadException(RcException):
    pass


class SSHException(RcException):
    pass


class MachineNotRunningException(RcException):
    pass


class RunException(RcException):
    pass


class MachineChangeTypeException(RcException):
    pass


class MachineNotReadyException(RcException):
    pass


class SaveImageException(RcException):
    pass


class DeleteImageException(RcException):
    pass


class FirewallRuleCreationException(RcException):
    pass


class FirewallRuleDeleteionException(RcException):
    pass


class DiskCreationException(RcException):
    pass


class DiskDeletionException(RcException):
    pass


class MachineAddDiskException(RcException):
    pass


class MachineRemoveDiskException(RcException):
    pass


class PmapException(RcException):
    pass



import subprocess
from rc.exception import RunException, PmapException
import sys
from collections import namedtuple
import os
from typing import Union, List
import io
from threading import Thread, Lock
from queue import Queue
from concurrent.futures import ThreadPoolExecutor, as_completed
import signal

_RunResult = namedtuple('_RunResult', ['stdout', 'stderr', 'returncode'])


class RunResult(_RunResult):
    @property
    def exitcode(self):
        return self.returncode


def convert_list_command_to_str(cmd: List[str]) -> str:
    cmd_str = io.StringIO('')
    for c in cmd:
        if c and (c[0] in ['"', "'", '<', '|', '>', '&', ';', '~']):
            cmd_str.write(c)
        elif len(c) >= 2 and c[:2] in ['1>', '2>']:
            cmd_str.write(c)
        else:
            cmd_str.write('"' + c + '"')
        cmd_str.write(' ')
    return cmd_str.getvalue()


def run(cmd: Union[str, List[str]], *, shell=['/bin/sh', '-c'], input=None, timeout=None, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
    p = running(cmd, shell=shell, input=input,
                text=text, stdout=stdout, stderr=stderr)
    try:
        stdout, stderr = p.communicate(timeout=timeout)
    except Exception as e:
        raise RunException(e) from None
    return RunResult(returncode=p.returncode, stdout=stdout, stderr=stderr)


def bash(script, *, timeout=None, flag='set -euo pipefail', login=False, interactive=False, run_shell=['/bin/sh', '-c']):
    cmd = 'bash '
    if login:
        cmd += '-l '
    if interactive:
        cmd += '-i '
    if flag:
        script = flag + '\n' + script

    return run(cmd, input=script, timeout=timeout, shell=run_shell)


def sudo(script, *, shell=None, user='root', timeout=None, flag='set -euo pipefail', run_shell=['/bin/sh', '-c']):
    cmd = 'sudo '
    if user != 'root':
        cmd += '-u ' + user + ' '
    if shell is None:
        cmd += '-i '
    else:
        cmd += '-s ' + shell + ' '
    if flag:
        script = flag + '\n' + script
    return run(cmd, input=script, timeout=timeout, shell=run_shell)


def python(script, *, timeout=None, python_path='python', user=None, run_shell=['/bin/sh', '-c']):
    if user:
        return sudo(script, user=user, flag='', shell=python_path, timeout=timeout, run_shell=run_shell)
    else:
        return run(python_path, input=script, timeout=timeout, shell=run_shell)


def python2(script, **kwargs):
    return python(script, **kwargs)


def python3(script, **kwargs):
    return python(script, **kwargs)


def running(cmd: Union[str, List[str]], *, shell=['/bin/sh', '-c'], input=None, text=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE):
    if type(cmd) is list:
        cmd = convert_list_command_to_str(cmd)
    try:
        if not shell:
            shell = []
        p = subprocess.Popen([*shell, cmd], stdin=subprocess.PIPE if input else None,
                             stdout=stdout, stderr=stderr,
                             universal_newlines=text, preexec_fn=os.setsid)
        if input:
            p.stdin.write(input)
            p.stdin.close()
            p.stdin = None
        return p
    except:
        e = sys.exc_info()
        # print(e[0].__name__)
        # print(e[1])
        # import traceback
        # traceback.print_exception(e[2])
        raise RunException(e[1]) from None


STDOUT = 1
STDERR = 2
EXIT = 3


def run_stream(cmd: Union[str, List[str]], *, shell=['/bin/sh', '-c'], input=None, text=True):
    p = running(cmd, shell=shell, input=input, text=text)
    q = Queue()

    def queue_stdout():
        for line in p.stdout:
            q.put((STDOUT, line))

    def queue_stderr():
        for line in p.stderr:
            q.put((STDERR, line))

    queue_stdout_thread = Thread(target=queue_stdout)
    queue_stderr_thread = Thread(target=queue_stderr)

    queue_stdout_thread.start()
    queue_stderr_thread.start()

    def queue_exit():
        queue_stdout_thread.join()
        queue_stderr_thread.join()

        while p.poll() is None:
            pass
        q.put((EXIT, p.returncode))

    queue_exit_thread = Thread(target=queue_exit)
    queue_exit_thread.start()
    return q, p


def handle_stream(q, *, stdout_handler=None, stderr_handler=None, exit_handler=None):
    while True:
        event, value = q.get()
        if event == EXIT:
            if exit_handler:
                exit_handler(value)
            break
        elif event == STDOUT:
            if stdout_handler:
                stdout_handler(value)
        elif event == STDERR:
            if stderr_handler:
                stderr_handler(value)
        q.task_done()


executor = ThreadPoolExecutor()


def go(func, *args, **kwargs):
    return executor.submit(func, *args, **kwargs)


def pmap(func, *iterables, timeout=None, on_exception='raise'):
    if on_exception == 'raise':
        return list(executor.map(func, *iterables, timeout=timeout))

    def func_wrapper(arg):
        try:
            return func(arg)
        except Exception as e:
            return PmapException(e, arg)
    return list(executor.map(func_wrapper, *iterables, timeout=timeout))


def print_stream(q, *, prefix):
    return handle_stream(q, stdout_handler=lambda line: print(prefix, 'STDOUT |', line, end=''),
                         stderr_handler=lambda line: print(
                             prefix, 'STDERR |', line, end=''),
                         exit_handler=lambda exitcode: print(prefix, 'EXIT CODE |', exitcode, end=''))


def save_stream_to_file(q, *, path, name):
    with open(os.path.join(path, name+'.stdout'), 'w') as stdout:
        with open(os.path.join(path, name+'.stderr'), 'w') as stderr:
            with open(os.path.join(path, name+'.exitcode'), 'w') as exitcode:
                return handle_stream(q, stdout_handler=lambda line: stdout.write(line),
                                     stderr_handler=lambda line: stderr.write(
                                         line),
                                     exit_handler=lambda ec: exitcode.write(ec))


print_lock = Lock()


def p(*args):
    with print_lock:
        print(*args, flush=True)


def ep(*args):
    with print_lock:
        print(*args, file=sys.stderr, flush=True)


def kill(p):
    os.killpg(os.getpgid(p.pid), signal.SIGTERM)


def ok(p):
    if p.returncode != 0:
        raise RunException(p.stderr)
    return p



class Firewall:
    def __init__(self, *, provider, name, direction='in', action='allow', ports, ips=['0.0.0.0/0']):
        self.provider = provider
        self.name = name
        self.direction = direction
        self.action = action
        self.ports = ports
        self.ips = ips

    def __eq__(self, other):
        return (self.provider, self.name, self.direction, self.action, self.ports, self.ips) == \
            (other.provider, other.name, other.action,
             other.ports, other.ips)

    def __expr__(self):
        if 'direction' == 'in':
            return f'Firewall(provider={self.provider}, name={self.name}, direction={self.direction}, action={self.action}, ports={self.ports}, ips={self.ips})'
        else:
            return f'Firewall(provider={self.provider}, name={self.name}, direction={self.direction}, action={self.action}, ports={self.ports}, ips={self.ips})'

    def machines(self):
        return self.provider.firewall_machines(self)

    def delete(self):
        return self.provider.delete_firewall(self)

    def add_to_machine(self, machine):
        return self.provider.add_firewall(machine, self)

    def remove_from_machine(self, machine):
        return self.provider.remove_firewall(machine, self)



from rc.exception import UploadException, DownloadException, MachineNotReadyException
from io import StringIO
from rc.util import run, run_stream, convert_list_command_to_str, \
    bash, sudo, python, python2, python3, running, kill, ok
from retry import retry
import datetime
import subprocess


class Machine:
    def __init__(self, *, provider, name, ip, username, ssh_key_path, **kwargs):
        self.provider = provider
        self.name = name
        self.ip = ip
        self.username = username
        self.ssh_key_path = ssh_key_path
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, other):
        return (self.provider, self.name, self.zone, self.ip, self.username, self.ssh_key_path) == (other.provider, other.name, other.zone, other.ip, other.username, other.ssh_key_path)

    def __repr__(self):
        return 'Machine(provider={provider}, name={name}, zone={zone}, ip={ip}, username={username}, ssh_key_path={ssh_key_path})'.format(
            provider=self.provider,
            name=self.name,
            zone=self.zone,
            ip=self.ip,
            username=self.username,
            ssh_key_path=self.ssh_key_path
        )

    def __str__(self):
        return f"{self.provider.__name__.split('.')[-1]}." + self.name

    def status(self):
        return self.provider.status(self)

    def bootup(self):
        return self.provider.bootup(self)

    def shutdown(self):
        return self.provider.shutdown(self)

    def delete(self):
        return self.provider.delete(self)

    def change_type(self):
        return self.provider.change_type(self)

    def upload(self, local_path, machine_path, switch_user=None, su=None, user=None):
        user = user or su or switch_user
        if user:
            rsync = f'''--rsync-path='sudo -u {user} rsync' '''
        else:
            rsync = ''

        p = run(f'''
rsync -e 'ssh -o StrictHostKeyChecking=no -i {self.ssh_key_path}' -r \
    {rsync} --progress {local_path} {self.username}@{self.ip}:{machine_path}
''')
        if p.returncode != 0:
            raise UploadException(p.stderr)
        return p

    def download(self, machine_path, local_path, sudo=True):
        if sudo:
            rsync = '''--rsync-path='sudo rsync' '''
        else:
            rsync = ''
        p = run(f'''
rsync -e 'ssh -o StrictHostKeyChecking=no -i {self.ssh_key_path}' -r \
    {rsync} --progress {self.username}@{self.ip}:{machine_path} {local_path}
''')
        if p.returncode != 0:
            raise DownloadException(p.stderr)
        return p

    def _ssh_shell(self):
        return ['ssh', '-o', 'StrictHostKeyChecking=no',
                *(['-i', self.ssh_key_path] if self.ssh_key_path else []), self.username + '@' + self.ip, '--']

    def ssh_shell_str(self):
        return ' '.join(self._ssh_shell()[:-1])

    def running(self, cmd, *, input=None):
        return running(cmd, shell=self._ssh_shell(), input=input)

    def run(self, cmd, *, timeout=None, input=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
        return run(cmd, shell=self._ssh_shell(), timeout=timeout, input=input, stdout=stdout, stderr=stderr)

    def run_stream(self, cmd, input=None):
        return run_stream(cmd, shell=self._ssh_shell(), input=input)

    def sudo(self, script, *, shell=None, user='root', timeout=None, flag='set -euo pipefail'):
        return sudo(script, shell=shell, user=user, timeout=timeout, flag=flag,
                    run_shell=self._ssh_shell())

    def bash(self, script, *, timeout=None, flag='set -euo pipefail', login=False, interactive=False):
        return bash(script, timeout=timeout, login=login, interactive=interactive, flag=flag,
                    run_shell=self._ssh_shell())

    def run_bg(self, script, *, cmd='bash', stdout='/tmp/python-rc.log', stderr='/tmp/python-rc.log', exitcode='/tmp/python-rc.exitcode', pid='/tmp/python-rc.pid', user=None):
        ts = datetime.datetime.strftime(
            datetime.datetime.utcnow(), '%Y%m%d_%H%M%S')
        ok(self.ensure_dir(f'/tmp/python-rc'))
        ok(self.edit(f'/tmp/python-rc/script_{ts}', script, user=user))
        print('here we go', datetime.datetime.now())
        run_cmd = f'( {cmd} /tmp/python-rc/script_{ts} >>{stdout} 2>>{stderr} & echo -n $! > {pid}; wait $!; echo -n $? >{exitcode}) </dev/null >/dev/null 2>/dev/null &'
        if user:
            p = ok(self.sudo(run_cmd, user=user))
        else:
            p = ok(self.run(run_cmd))
        print(p.stdout)
        print('aaaaa', datetime.datetime.now())

    def python(self, script, *, timeout=None, python_path='python', user=None):
        return python(script, timeout=timeout, python_path=python_path, user=user, run_shell=self._ssh_shell())

    def python2(self, script, **kwargs):
        return python2(script, run_shell=self._ssh_shell(), **kwargs)

    def python3(self, script, **kwargs):
        return python3(script, run_shell=self._ssh_shell(), **kwargs)

    def run_detach_tmux(self, cmd: str, *, name='python-rc', log='/tmp/python-rc.log'):
        return self.run(f"tmux new -s {name} -d '{cmd}' \\; pipe-pane 'cat > {log}'")

    def kill_detach_tmux(self, name='python-rc'):
        return self.run(f"tmux kill-session -t {name}")

    def save_image(self, image, **kwargs):
        return self.provider.save_image(self, image, **kwargs)

    @retry(MachineNotReadyException)
    def wait_ssh(self):
        p = self.run('echo a')
        if p.returncode != 0:
            raise MachineNotReadyException(p.stderr)

    def firewalls(self):
        return self.provider.machine_firewalls(self)

    def ensure_user(self, username, *, sudo=True, pubkey=True):
        add_sudo = ''
        if sudo:
            add_sudo = f'echo "{username}  ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/{username}'
        add_pubkey = ''
        if username == 'root':
            ssh_path = '/root/.ssh'
        else:
            ssh_path = f'/home/{username}/.ssh'
        if pubkey is True:
            pubkey = run(
                f'ssh-keygen -y -f {self.ssh_key_path}').stdout.strip()
        if pubkey:
            add_pubkey = f'''sudo -u {username} mkdir -p {ssh_path}
            if ! $(grep -Fxq '{pubkey}' {ssh_path}/authorized_keys); then
                sudo -u {username} echo "{pubkey}" >> {ssh_path}/authorized_keys
            fi
            '''
        cmd = f'''
        useradd {username} --create-home --shell /bin/bash || true
        {add_sudo}
        {add_pubkey}
        '''
        p = self.sudo(cmd)
        return p

    def edit(self, path, content, append=False, user=None):
        if append:
            op = '>>'
        else:
            op = '>'
        cmd = f"""cat {op} {path} <<'c7a88caeb23f4ac0f377c59b703fb7f1091d0708'
{content}
c7a88caeb23f4ac0f377c59b703fb7f1091d0708"""
        if user:
            return self.sudo(cmd, user=user)
        else:
            return self.bash(cmd)

    def ensure_lines(self, path, lines, user=None):
        pass

    def read(self, path):
        pass

    def ensure_file(self):
        pass

    def ensure_dir(self, path, user=None, allow_write=True):
        cmd = f'mkdir -p {path}'
        if allow_write:
            cmd += f'\nchmod a+w {path}'
        if user:
            return self.sudo(cmd, user=user)
        else:
            return self.bash(cmd)



from rc.util import run
from rc.exception import MachineCreationException, MachineDeletionException, \
    MachineShutdownException, MachineBootupException, SaveImageException
from rc.machine import Machine
import json
import sys
import os

azure_provider = sys.modules[__name__]

SSH_KEY_PATH = os.path.expanduser('~/.ssh/id_rsa')


def _group_exist(group_name):
    p = run(['az', 'group', 'exists', '-n', group_name])
    if p.stdout.strip() == 'true':
        return True
    return False


def _create_group(group_name, location):
    return run(['az', 'group', 'create', '-n', group_name, '-l', location])


def _delete_group(group_name):
    # Very slow
    return run(['az', 'group', 'delete', '-n', group_name, '--yes'])


def create(*, name, machine_size, disk_size_gb=None, image, location):
    args = ['--name', name]
    args += ['--resource-group', name]
    args += ['--image', image]
    args += ['--size', machine_size]

    if disk_size_gb is not None:
        args += ['--os-disk-size-gb', str(disk_size_gb)]
    if _group_exist(name):
        raise MachineCreationException('resource group already exist')

    p = _create_group(name, location)
    if p.returncode != 0:
        raise MachineCreationException(p.stderr)

    p = run(['az', 'vm', 'create', *args])
    if p.returncode != 0:
        _delete_group(name)
        raise MachineCreationException(p.stderr)

    ip = json.loads(p.stdout)['publicIpAddress']

    machine = Machine(provider=azure_provider, name=name, zone=None,
                      ip=ip, username=os.getlogin(), ssh_key_path=SSH_KEY_PATH)
    return machine


def delete(machine):
    p = _delete_group(machine.name)
    if p.returncode != 0:
        raise MachineDeletionException(p.stderr)


def get(name):
    p = run(['az', 'vm', 'list-ip-addresses', '-n', name, '-g', name])
    if p.returncode == 0:
        res = json.loads(p.stdout)
        if not res:
            return None
        ip_address = res[
            0]['virtualMachine']['network']['publicIpAddresses'][0]['ipAddress']
        return Machine(provider=azure_provider, name=name, zone=None,
                       ip=ip_address, username=os.getlogin(), ssh_key_path=SSH_KEY_PATH)
    else:
        return None


def shutdown(machine):
    p = run(['az', 'vm', 'stop',
             '-n', machine.name, '-g', machine.name])
    if p.returncode != 0:
        raise MachineShutdownException(p.stderr)


def bootup(machine):
    p = run(['az', 'vm', 'start',
             '-n', machine.name, '-g', machine.name])
    if p.returncode != 0:
        raise MachineBootupException(p.stderr)


def save_image(machine, image, *, group):
    p = run(['az', 'image', 'create', '-n', image, '-g',
             group if group else machine.name, '--source', machine.name])
    if p.returncode != 0:
        raise SaveImageException(p.stderr)



from rc.cli import config
from rc import run, pmap, p, ep, RunResult, RunException
from rc.util import convert_list_command_to_str
import subprocess
import os
import libtmux
import argparse
from pytimeparse.timeparse import timeparse
from threading import Lock
import term


def execute(targets, args):
    execute_argparser = argparse.ArgumentParser(
        'execute a command on target of machines')
    execute_argparser.add_argument(
        '-t', '--timeout', help='timeout to execute commands. ', default='2m')
    execute_argparser.add_argument('command')
    execute_argparser.add_argument('args', nargs=argparse.REMAINDER)
    args = execute_argparser.parse_args(args)
    timeout = timeparse(args.timeout)
    if args.args:
        cmd = convert_list_command_to_str([args.command, *args.args])
    else:
        cmd = args.command

    l = Lock()
    for target in targets:
        print(f'Start executing on {target}')

    def exec(i):
        target = targets[i]
        ret = None
        try:
            log_path = os.path.join(config.logs_dir, str(target))
            log = open(log_path, 'w')
            proc = target.run(cmd, timeout=timeout, stdout=log, stderr=log)
            if proc.returncode == 0:
                output = f'{term.green}SUCCESS{term.off} on {target}'
            else:
                output = f'{term.red}FAIL{term.off} on {target}. Exit code: {proc.returncode}'
            ret = proc
        except RunException as e:
            output = f'{term.red}FAIL{term.off} on {target}. Timeout'
            ret = e
        output += f'. Log: file://{log_path}'
        with l:
            term.saveCursor()
            term.up(len(targets) - i)
            term.clearLine()
            term.writeLine(output)
            term.restoreCursor()
        return ret
    results = pmap(exec, range(len(targets)))
    if all(map(lambda r: isinstance(r, RunResult) and r.returncode == 0, results)):
        term.writeLine('All execution succeeded', term.green)
        exit(0)
    else:
        term.writeLine('Some execution failed', term.red)
        exit(1)


def cat(args):
    group_name = args[0]
    group_config_file = config.get(group_name)
    if group_config_file:
        p(open(group_config_file).read())
    else:
        p(f'group {group_name} not exists')
        exit(2)


def ls(_args):
    p(run(f'ls {config.groups_dir}').stdout)


def edit(args):
    group_name = args[0]
    group_config_file = os.path.join(config.groups_dir, group_name)
    subprocess.run(['vi', group_config_file])


def rm(args):
    group_name = args[0]
    os.remove(os.path.join(config.groups_dir, group_name))


def tmux(args):
    targets = config.get_targets(args[0])
    if not targets:
        ep(f'targets not exist: {args[0]}')
        exit(2)

    server = libtmux.Server(socket_name='python-rc', config_file=os.path.join(
        config.config_dir, '.tmux.conf'))
    try:
        sessions = server.list_sessions()
    except:
        sessions = []
    max_python_sessions = -1
    for s in sessions:
        if s.name.startswith('python-rc'):
            n = int(s.name.split('-')[-1])
            max_python_sessions = max(max_python_sessions, n)
    max_python_sessions += 1

    session_name = f"python-rc-{max_python_sessions}"
    session = server.new_session(session_name)

    n = len(targets)
    w = session.attached_window.id
    window = session.new_window()
    pane = window.panes[0]
    session.kill_window(w)
    for i in range(1, n):
        window.split_window()
    window.select_layout('tiled')
    for i, pane in enumerate(window.panes):
        pane.send_keys('\n')
        pane.send_keys(targets[i].ssh_shell_str())
        pane.send_keys('\n')
    window.set_window_option('synchronize-panes', 'on')
    window.panes[0].select_pane()
    subprocess.run(['tmux', '-L', 'python-rc', 'a', '-t', session_name])



import re
from functools import lru_cache
import yaml
import os
from retry import retry
import sys
from rc.machine import Machine
from rc.util import run
from rc.exception import MachineCreationException, MachineNotRunningException, MachineShutdownException, \
    MachineDeletionException, MachineChangeTypeException, MachineNotReadyException, SaveImageException, \
    DeleteImageException, FirewallRuleCreationException, FirewallRuleDeleteionException, MachineBootupException, \
    RcException, DiskCreationException, DiskDeletionException, MachineAddDiskException, MachineRemoveDiskException

gcloud_provider = sys.modules[__name__]


def _zone_region(zone):
    return zone[:-2]


def create_firewall(name, allows=None, *, project=None):
    if allows:
        allow_param = ['--allow', ','.join(allows)]
    else:
        allow_param = ['--allow', 'tcp:22']

    cmd = ['gcloud', 'compute', 'firewall-rules', 'create',
           name, '--target-tags', name, *allow_param]
    if project:
        cmd += ['--project', project]
    return run(cmd)


def _reserve_ip_address(ip, name, region, *, project=None):
    cmd = ['gcloud', 'compute', 'addresses',
           'create', name, '--addresses', ip, '--region', region]
    if project:
        cmd += ['--project', project]
    return run(cmd)


def delete_firewall(name, *, project=None):
    cmd = ['gcloud', 'compute', 'firewall-rules', 'delete',
                name]
    if project:
        cmd += ['--project', project]
    return run(cmd, input='yes\n')


def _address_exist(name, region, *, project=None):
    cmd = ['gcloud', 'compute', 'addresses', 'describe', '--region', region, name]
    if project:
        cmd += ['--project', project]
    return run(cmd).returncode == 0


def firewall_exist(name, *, project=None):
    cmd = ['gcloud', 'compute', 'firewall-rules', 'describe', name]
    if project:
        cmd += ['--project', project]
    return run(cmd).returncode == 0


def _release_ip_address(name, region, *, project=None):
    cmd = ['gcloud', 'compute', 'addresses', 'delete', '--region', region, name]
    if project:
        cmd += ['--project', project]
    return run(cmd, input='yes\n')


SSH_KEY_PATH = os.path.expanduser('~/.ssh/google_compute_engine')

if not os.path.exists(SSH_KEY_PATH):
    print("There is no key at {}, creating new key.".format(SSH_KEY_PATH))
    # Generate new ssh key to log in.
    # WARNING: It will create a key with an empty passphrase.
    run(["ssh-keygen", "-f", SSH_KEY_PATH, "-t", "rsa", "-N", '""'])
    # Upload key
    run(["gcloud", "compute", "os-login", "ssh-keys",
         "add", "--key-file={}.pub".format(SSH_KEY_PATH)])


def list(*, pattern=None, project=None, username=None, ssh_key_path=None):
    cmd = ['gcloud', 'compute', 'instances', 'list', '--format',
           'value(zone, name, networkInterfaces[0].accessConfigs[0].natIP)']
    if project:
        cmd += ['--project', project]
    else:
        project = get_project()
    p = run(cmd)
    result = []
    lines = p.stdout.strip('\n').split('\n')
    for line in lines:
        zone, name, ip = re.split(r'\s+', line)
        if pattern:
            if pattern not in name:
                continue
        if username is None:
            username = _get_username(project)
        if ssh_key_path is None:
            ssh_key_path = SSH_KEY_PATH
        result.append(Machine(provider=gcloud_provider, name=name,
                              zone=zone, ip=ip, username=username, ssh_key_path=ssh_key_path,
                              project=project))
    return result


def get(name, *, username=None, ssh_key_path=None, project=None, **kwargs):
    cmd = ['gcloud', 'compute', 'instances', 'list', '--format',
             'value(zone, name, networkInterfaces[0].accessConfigs[0].natIP)',
             '--filter', 'name=' + name]
    if project:
        cmd += ['--project', project]
    else:
        project = get_project()
    p = run(cmd)
    if p.stdout.strip('\n') == '':
        return None
    zone, name, ip = re.split(r'\s+', p.stdout.strip('\n'))
    if username is None:
        username = _get_username(project)
    if ssh_key_path is None:
        ssh_key_path = SSH_KEY_PATH
    return Machine(provider=gcloud_provider, name=name,
                   zone=zone, ip=ip, username=username, ssh_key_path=ssh_key_path,
                   project=project)


@lru_cache()
def _get_username(project):
    p = run(f'gcloud compute project-info describe --project {project}')
    project_metadata = yaml.safe_load(p.stdout)['commonInstanceMetadata']['items']
    enable_oslogin = False
    for md in project_metadata:
        if md['key'] == 'enable-oslogin':
            if md['value'] == 'TRUE':
                enable_oslogin = True
            break
    if enable_oslogin:
        p = run(['gcloud', 'compute', 'os-login', 'describe-profile'])
        return yaml.safe_load(p.stdout)['posixAccounts'][0]['username']
    else:
        return os.getlogin()


def _get_ip(name, *, project=None):
    cmd = ['gcloud', 'compute', 'instances', 'list', '--filter', 'name='+name, '--format',
             'get(networkInterfaces[0].accessConfigs[0].natIP)']
    if project:
        cmd += ['--project', project]
    p = run(cmd)
    return p.stdout.strip()


@retry(MachineNotRunningException, delay=2)
def _wait_bootup(name, *, project=None):
    cmd = ['gcloud', 'compute', 'instances', 'list', '--format',
             'value(status)', '--filter', 'name=' + name]
    if project:
        cmd += ['--project', project]
    p = run(cmd)
    status = p.stdout.strip()
    if status != 'RUNNING':
        raise MachineNotRunningException(status)


def create(name, *, project=None, machine_type, disk_size, image_project, image_family=None, image=None, zone, min_cpu_platform=None,
           preemptible=False, firewall_allows=None, reserve_ip=True, firewalls=None, disk_type=None):
    args = [name]
    args += ['--machine-type', machine_type]
    args += ['--boot-disk-size', disk_size]
    if disk_type:
        args += ['--boot-disk-type', disk_type]
    if image_family:
        args += ['--image-family', image_family]
    args += ['--image-project', image_project]
    if image:
        args += ['--image', image]
    args += ['--zone', zone]

    if min_cpu_platform:
        args += ['--min-cpu-platform', min_cpu_platform]
    if preemptible:
        args += ['--preemptible']

    if firewall_allows:
        if firewall_exist(name, project=project):
            delete_firewall(name, project=project)
        p = create_firewall(name=name, allows=firewall_allows, project=project)
        if p.returncode != 0:
            raise MachineCreationException(p.stderr)
        args += ['--tags', name]
    elif firewalls:
        args += ['--tags', ','.join(firewalls)]
    if project:
        args += ['--project', project]

    p = run(['gcloud', 'compute', 'instances', 'create', *args])
    if p.returncode != 0:
        if firewall_allows:
            delete_firewall(name, project=project)
        raise MachineCreationException(p.stderr)

    _wait_bootup(name, project=project)

    ip = _get_ip(name, project=project)
    if not preemptible and reserve_ip:
        if _address_exist(name, _zone_region(zone), project=project):
            _release_ip_address(name, _zone_region(zone), project=project)
        p = _reserve_ip_address(ip, name, _zone_region(zone), project=project)
        if p.returncode != 0:
            if firewall_allows:
                delete_firewall(name, project=project)
            _delete_machine(name, zone, project=project)
            raise MachineCreationException(p.stderr)
    if not project:
        project = get_project()
    machine = Machine(provider=gcloud_provider, name=name, zone=zone,
                      ip=ip, username=_get_username(project), ssh_key_path=SSH_KEY_PATH,
                      project=project)
    machine.wait_ssh()
    return machine


def delete(machine):
    project = machine.project
    p = _delete_machine(machine.name, machine.zone, project=project)
    if p.returncode != 0:
        raise MachineDeletionException(p.stderr)
    if _address_exist(machine.name, _zone_region(machine.zone), project=project):
        p = _release_ip_address(machine.name, _zone_region(machine.zone), project=project)
        if p.returncode != 0:
            raise MachineDeletionException(p.stderr)
    if firewall_exist(machine.name, project=project):
        p = delete_firewall(machine.name, project=project)
        if p.returncode != 0:
            raise MachineDeletionException(p.stderr)


def _delete_machine(name, zone, *, project=None):
    cmd = ['gcloud', 'compute', 'instances',
                'delete', name, '--zone', zone]
    if project:
        cmd += ['--project', project]
    return run(cmd, input='yes\n')


def shutdown(machine):
    p = run(['gcloud', 'compute', 'instances', 'stop',
             machine.name, '--zone', machine.zone, '--project', machine.project])
    if p.returncode != 0:
        raise MachineShutdownException(p.stderr)


def bootup(machine):
    p = run(['gcloud', 'compute', 'instances', 'start',
             machine.name, '--zone', machine.zone, '--project', machine.project])
    if p.returncode != 0:
        raise MachineBootupException(p.stderr)
    machine.wait_ssh()


def status(machine):
    p = run(['gcloud', 'compute', 'instances', 'list', '--format',
             'value(status)', '--filter', 'name=' + machine.name, '--project', machine.project])
    return p.stdout.strip()


def change_type(machine, new_type):
    p = run(['gcloud', 'compute', 'instances', 'set-machine-type', machine.name,
             '--zone', machine.zone, '--machine-type', new_type, '--project', machine.project])
    if p.returncode != 0:
        raise MachineChangeTypeException(p.stderr)


def save_image(machine, image, *, image_family=None, description=None):
    command = ['gcloud', 'compute', 'images', 'create', image, '--source-disk',
               machine.name, '--source-disk-zone', machine.zone, '--project', machine.project]
    if image_family:
        command += ['--family', image_family]
    if description:
        command += ['--description', description]
    p = run(command)
    if p.returncode != 0:
        raise SaveImageException(p.stderr)


def delete_image(image, *, project=None):
    command = ['gcloud', 'compute', 'images', 'delete', image]
    if project:
        command += ['--project', project]
    p = run(command)
    if p.returncode != 0:
        raise DeleteImageException(p.stderr)


def add_firewall(machine, firewall):
    pass


def remove_firewall(machine, firewall):
    pass


def create_disk(name, *, type='pd-standard', size, project=None, zone):
    if project:
        project = f'--project {project}'
    else:
        project = ''
    p = run(
        f'gcloud beta compute disks create {name} {project} --type={type} --size={size} --zone={zone}')
    if p.returncode != 0:
        raise DiskCreationException(p.stderr)


def add_disk(machine, disk):
    p = run(
        f'gcloud compute instances attach-disk {machine.name} --disk={disk} --zone {machine.zone} --device-name={disk} --project {machine.project}')
    if p.returncode != 0:
        raise MachineAddDiskException(p.stderr)
    p = machine.sudo('lsblk -f /dev/disk/by-id/google-' +
                     disk + " | awk '{print $2}'")
    if p.returncode != 0:
        raise MachineAddDiskException(p.stderr)
    fstype = p.stdout.strip()
    assert fstype.startswith('FSTYPE')
    if fstype.strip() != 'FSTYPE':
        # it's already been formatted and there's a filesystem on it
        p = machine.sudo(f'''
        mkdir -p /mnt/{disk}
        mount -o discard,defaults /dev/disk/by-id/google-{disk} /mnt/{disk}
        chmod a+w /mnt/{disk}
        ''')
    else:
        p = machine.sudo(f'''
        sudo mkfs.ext4 -m 0 -F -E lazy_itable_init=0,lazy_journal_init=0,discard /dev/disk/by-id/google-{disk}
        mkdir -p /mnt/{disk}
        mount -o discard,defaults /dev/disk/by-id/google-{disk} /mnt/{disk}
        sudo chmod a+w /mnt/{disk}
        ''')
    if p.returncode != 0:
        raise MachineAddDiskException(p.stderr)


def remove_disk(machine, disk):
    p = machine.sudo(f'umount /dev/disk/by-id/google-{disk}')
    if p.returncode != 0 and 'No such file or directory' not in p.stderr:
        raise MachineRemoveDiskException(p.stderr)
    p = run(
        f'gcloud compute instances detach-disk {machine.name} --disk {disk} --zone {machine.zone} --project {machine.project}'
    )
    if p.returncode != 0:
        raise MachineRemoveDiskException(p.stderr)


def delete_disk(disk, zone, *, project=None):
    cmd = f'gcloud compute disks delete {disk} --zone {zone}'
    if project:
        cmd += f' --project {project}'
    p = run(cmd, input = 'y\n')
    if p.returncode != 0:
        raise DiskDeletionException(p.stderr)


def set_project(project):
    p = run(f'gcloud config set project {project}')
    if p.returncode != 0:
        raise RcException(p.stderr)


def get_project():
    p = run(f'gcloud config get-value project')
    if p.returncode != 0:
        raise RcException(p.stderr)
    return p.stdout.strip()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="python-rc",  # Replace with your own username
    version="0.3.7",
    author="Bo Yao",
    author_email="icerove@gmail.com",
    description="Python remote control library for programmatically control remote machines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ailisp/python-rc",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Systems Administration"
    ],
    install_requires=[
        'retry',
        'PyYAML',
        'humanfriendly',
        'libtmux',
        'py-term',
        'pytimeparse',
    ],
    python_requires='>=3.6',
    scripts=['bin/rc']
)



from timeit import default_timer as timer
from humanfriendly import format_timespan


class Timer():
    def __init__(self, action):
        self.action = action

    def __enter__(self):
        self.start = timer()

    def __exit__(self, type, value, traceback):
        print("Time to", self.action+":", format_timespan(timer()-self.start))



from rc import azure
from rc.util import run
from rc.test.util import Timer


def create_test_machine():
    return azure.create(name='test-rc-node', machine_size='Standard_D1', image='ubuntults', location='westus')


def test_azure():
    # Cleanup from last fail
    a = azure.get('test-rc-node')
    print(a)
    if a is not None:
        a.delete()

    # Test create machine
    with Timer('create machine'):
        machine1 = create_test_machine()
    assert machine1.name == "test-rc-node"
    assert machine1.ip != ''
    assert machine1.ssh_key_path != ''
    assert machine1.username != ''

    # Test get machine
    with Timer('get machine'):
        machine1_get = azure.get('test-rc-node')
    assert machine1 == machine1_get

    # Test delete
    with Timer('delete machine'):
        machine1.delete()
    assert azure.get('test-rc-node') is None
